__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from .document import DocumentSet
from .querylang import QueryLangSet
from .chunk import ChunkSet
from .match import MatchSet
